@@ -284,6 +284,17 @@ void SpectacleCore::screenshotUpdated(const QPixmap &thePixmap)
         break;
     case StartMode::Gui:
         mMainWindow->setScreenshotAndShow(thePixmap);
+
+        bool autoSaveImage = SpectacleConfig::instance()->autoSaveImage();
+        bool copyImageToClipboard = SpectacleConfig::instance()->copyImageToClipboard();
+
+        if (autoSaveImage && copyImageToClipboard) {
+            lExportManager->doSaveAndCopy();
+        } else if (autoSaveImage) {
+            lExportManager->doSave();
+        } else if (copyImageToClipboard) {
+            lExportManager->doCopyToClipboard(false);
+        }
     }
 }
 
